"""
PDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

    Maintain and explore the contents of nested collections.

"""

import asyncio

# import json
from pyegeria._client import Client
from pyegeria.core._globals import max_paging_size, NO_ELEMENTS_FOUND


class ValidMetadataManager(Client):
    """The Valid Metadata OMVS provides APIs for retrieving and updating lists of valid metadata values.
        For more details see: https://egeria-project.org/guides/planning/valid-values/overview/

    Attributes:

        view_server: str
            The name of the View Server to connect to.
        platform_url : str
            URL of the server platform to connect to
        user_id : str
            The identity of the user calling the method - this sets a default optionally used by the methods
            when the user doesn't pass the user_id on a method call.
        user_pwd: str
            The password associated with the user_id. Defaults to None

    """

    def __init__(
        self,
        view_server: str,
        platform_url: str,
        user_id: str = None,
        user_pwd: str = None,
        token: str = None,
    ):
        self.view_server = view_server
        self.platform_url = platform_url
        self.user_id = user_id
        self.user_pwd = user_pwd

        self.valid_m_command_base: str = f"/api/open-metadata/valid-metadata"
        self.page_size = max_paging_size
        Client.__init__(self, view_server, platform_url, user_id, user_pwd, token=token)

    async def _async_setup_valid_metadata_value(
        self, property_name: str, type_name: str, body: dict
    ):
        """Create or update the valid value for a particular open metadata property name. If the typeName is null,
        this valid value applies to properties of this name from all types. The valid value is stored in the
        preferredValue property. If a valid value is already set up for this property (with overlapping effective dates)
        then the valid value is updated. Async Version.

        Parameters
        ----------
        property_name : str
            The name of the property for which the valid metadata value is being set up.
        type_name : str
            The name of the type for the valid metadata value.
        body : dict
            The body of the request containing the details of the valid metadata value.


        Returns
        -------
        No value is returned.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action


        Notes
        -----

        Payload structure similar to:
        {
          "displayName": "",
          "description": "",
          "preferredValue": "",
          "dataType": "",
          "isCaseSensitive": false,
          "isDeprecated" : false,
          "effectiveFrom" : "2024-09-30T20:00:00.000Z",
          "effectiveTo" : "2025-09-30T20:00:00.000Z",
        }
        """

        url = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/valid-metadata/setup-value/{property_name}?"
            f"typeName={type_name}"
        )

        await self._async_make_request("POST", url, body)
        return

    def setup_valid_metadata_value(
        self, property_name: str, type_name: str, body: dict
    ):
        """Create or update the valid value for a particular open metadata property name. If the typeName is null,
        this valid value applies to properties of this name from all types. The valid value is stored in the
        preferredValue property. If a valid value is already set up for this property (with overlapping effective dates)
        then the valid value is updated.

        Parameters
        ----------
        property_name : str
            The name of the property for which the valid metadata value is being set up.
        type_name : str
            The name of the type for the valid metadata value.
        body : dict
            The body of the request containing the details of the valid metadata value.


        Returns
        -------
        None - this method does not return a value.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        Notes
        -----

        Payload structure similar to:
        {
          "displayName": "",
          "description": "",
          "preferredValue": "",
          "dataType": "",
          "isCaseSensitive": false,
          "isDeprecated" : false,
          "additionalProperties": {
            "colour": "purple"
            }
        }
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_setup_valid_metadata_value(property_name, type_name, body)
        )
        return

    async def _async_setup_valid_metadata_map_name(
        self, property_name: str, type_name: str, body: dict
    ):
        """Create or update the valid value for a name that can be stored in a particular open metadata property name.
        This property is of type map from name to string. The mapName is stored in the preferredValue property of
        validMetadataValue. If the typeName is null, this valid value applies to properties of this name from any
        open metadata type. If a valid value is already set up for this property (with overlapping effective dates)
        then the valid value is updated. Async Version.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        body : dict
            The metadata map setup data.


        Returns
        -------
        None
            This method does not return any value.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action


        Notes
        -----

        Body strycture similar to:

        {
          "displayName": "",
          "description": "",
          "preferredValue": "put mapName value here",
          "dataType": "",
          "isCaseSensitive": false,
          "isDeprecated" : false,
          "effectiveFrom" : "2024-09-30T20:00:00.000Z",
          "effectiveTo" : "2025-09-30T20:00:00.000Z"
        }

        """

        url = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/valid-metadata/setup-map-name/{property_name}?"
            f"typeName={type_name}"
        )

        await self._async_make_request("POST", url, body)
        return

    def setup_valid_metadata_map_name(
        self, property_name: str, type_name: str, body: dict
    ):
        """Create or update the valid value for a name that can be stored in a particular open metadata property name.
        This property is of type map from name to string. The mapName is stored in the preferredValue property of
        validMetadataValue. If the typeName is null, this valid value applies to properties of this name from any
        open metadata type. If a valid value is already set up for this property (with overlapping effective dates)
        then the valid value is updated.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        body : dict
            The metadata map setup data.


        Returns
        -------
        None
            This method does not return any value.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action


        Notes
        -----

        Body strycture similar to:

        {
          "displayName": "",
          "description": "",
          "preferredValue": "put mapName value here",
          "dataType": "",
          "isCaseSensitive": false,
          "isDeprecated" : false
        }

        """

        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_setup_valid_metadata_map_name(property_name, type_name, body)
        )
        return

    async def _async_setup_valid_metadata_map_value(
        self, property_name: str, type_name: str, map_name: str, body: dict
    ) -> None:
        """Create or update the valid value for a name that can be stored in a particular open metadata property name.
        This property is of type map from name to string.
        The valid value is stored in the preferredValue property of validMetadataValue.
        If the typeName is null, this valid value applies to properties of this name from any open metadata type.
        If a valid value is already set up for this property (with overlapping effective dates) then the valid value
        is updated.  Async version.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        map_name: str
            The name of a map we associate a value with.
        body : dict
            The metadata map setup data.

        Returns
        -------
        None
            This method does not return any value.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action


        Notes
        -----

        Body strycture similar to:

        {
          "displayName": "",
          "description": "",
          "preferredValue": "put mapName value here",
          "dataType": "",
          "isCaseSensitive": false,
          "effectiveFrom" : "2024-09-30T20:00:00.000Z",
          "effectiveTo" : "2025-09-30T20:00:00.000Z"
        }
        """
        url = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/valid-metadata/setup-map-value/"
            f"{property_name}/{map_name}?typeName={type_name}"
        )

        await self._async_make_request("POST", url, body)
        return

    def setup_valid_metadata_map_value(
        self, property_name: str, type_name: str, map_name: str, body: dict
    ) -> None:
        """Create or update the valid value for a name that can be stored in a particular open metadata property name.
        This property is of type map from name to string.
        The valid value is stored in the preferredValue property of validMetadataValue.
        If the typeName is null, this valid value applies to properties of this name from any open metadata type.
        If a valid value is already set up for this property (with overlapping effective dates) then the valid value
        is updated.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        map_name: str
            The name of a map we associate a value with.
        body : dict
            The metadata map setup data.


        Returns
        -------
        None
            This method does not return a value.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action


        Notes
        -----

        Body structure similar to:

        {
          "displayName": "",
          "description": "",
          "preferredValue": "put mapName value here",
          "dataType": "",
          "isCaseSensitive": false,
        }
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_setup_valid_metadata_map_value(
                property_name, type_name, map_name, body
            )
        )
        return

    async def _async_clear_valid_metadata_value(
        self, property_name: str, type_name: str, preferred_value: str
    ) -> None:
        """Remove a valid value for a property. Async version.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        preferred_value: str
            The reference valye to remove.

        Returns
        -------
        None - This method does not return a value.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action


        """
        url = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/valid-metadata/clear-value/"
            f"{property_name}?typeName={type_name}&preferredValue={preferred_value}"
        )

        await self._async_make_request("POST", url)
        return

    def clear_valid_metadata_value(
        self, property_name: str, type_name: str, preferred_value: str
    ) -> None:
        """Remove a valid value for a property.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        preferred_value: str
            The reference valye to remove.

        Returns
        -------
        None - This method does not return a value.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action


        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_clear_valid_metadata_value(
                property_name, type_name, preferred_value
            )
        )
        return

    async def _async_clear_valid_metadata_map_name(
        self,
        property_name: str,
        type_name: str,
        map_name: str,
    ):
        """Remove a valid map name value for a property. The match is done on MapName name. Async version.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        map_name: str
            The name of a map we associate a value with.

        Returns
        -------
        None - This method does not return a value.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        url = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/valid-metadata/clear-map-name/"
            f"{property_name}?typeName={type_name}&mapName={map_name}"
        )

        await self._async_make_request("POST", url)
        return

    def clear_valid_metadata_map_name(
        self,
        property_name: str,
        type_name: str,
        map_name: str,
    ):
        """Remove a valid map name value for a property. The match is done on MapName name.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        map_name: str
            The name of a map we associate a value with.

        Returns
        -------
        None - This method does not return a value.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_clear_valid_metadata_map_name(
                property_name, type_name, map_name
            )
        )
        return

    async def _async_clear_valid_metadata_map_value(
        self, property_name: str, type_name: str, preferred_value: str
    ):
        """Remove a valid map name value for a property.  The match is done on preferred name. Async version.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        preferred_value: str
            The value to remove.

        Returns
        -------
        None - This method does not return a value.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        url = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/valid-metadata/clear-map-value/"
            f"{property_name}?typeName={type_name}&preferredValue={preferred_value}"
        )

        await self._async_make_request("POST", url)
        return

    def clear_valid_metadata_map_value(
        self, property_name: str, type_name: str, preferred_value: str
    ):
        """Remove a valid map name value for a property.  The match is done on preferred name.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        preferred_value: str
            The value to remove.

        Returns
        -------
        None - This method does not return a value.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_clear_valid_metadata_map_value(
                property_name, type_name, preferred_value
            )
        )
        return

    async def _async_validate_metadata_value(
        self, property_name: str, type_name: str, actual_value: str
    ) -> bool | str:
        """Validate whether the value found in an open metadata property is valid. Async version.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        actual_value: str
            The value to validate.

        Returns
        -------
        Bool - True, if validated.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        url = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/valid-metadata/validate-value/"
            f"{property_name}?typeName={type_name}&actualValue={actual_value}"
        )

        response = await self._async_make_request("GET", url)
        return response.json().get("flag", "No flag found")

    def validate_metadata_value(
        self, property_name: str, type_name: str, actual_value: str
    ) -> bool | str:
        """Validate whether the value found in an open metadata property is valid.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        actual_value: str
            The value to validate.

        Returns
        -------
        Bool - True, if validated.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_clear_valid_metadata_map_value(
                property_name, type_name, actual_value
            )
        )
        return response

    async def _async_validate_metadata_map_name(
        self, property_name: str, type_name: str, map_name: str
    ) -> bool | str:
        """Validate whether the name found in an open metadata map property is valid. Async version.

         Parameters
         ----------
         property_name : str
             The name of the property to setup metadata map.
         type_name : str
             The type name of the property.
        map_name: str
             The name of a map to validate.

         Returns
         -------
         Bool - True, if validated.

         Raises
         ------
         PyegeriaInvalidParameterException
           If the client passes incorrect parameters on the request - such as bad URLs or invalid values
         PyegeriaAPIException
           Raised by the server when an issue arises in processing a valid request
         NotAuthorizedException
           The principle specified by the user_id does not have authorization for the requested action

        """

        url = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/valid-metadata/validate-map-name/"
            f"{property_name}?typeName={type_name}&mapName={map_name}"
        )

        response = await self._async_make_request("GET", url)
        return response.json().get("flag", "No flag found")

    def validate_metadata_map_name(
        self, property_name: str, type_name: str, map_name: str
    ) -> bool | str:
        """Validate whether the name found in an open metadata map property is valid.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        map_name: str
            The name of a map to validate.

        Returns
        -------
        Bool - True, if validated.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_validate_metadata_map_name(property_name, type_name, map_name)
        )
        return response

    async def _async_validate_metadata_map_value(
        self, property_name: str, type_name: str, map_name: str, actual_value: str
    ) -> bool | str:
        """Validate whether the name found in an open metadata map property is valid. Async version.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        map_name: str
            The name of a map to validate.
        actual_value: str
            The actual value associated with the map to validate.

        Returns
        -------
        Bool - True, if validated.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        url = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/valid-metadata/validate-map-value/"
            f"{property_name}/{map_name}?typeName={type_name}&actualValue={actual_value}"
        )

        response = await self._async_make_request("GET", url)
        return response.json().get("flag", "No flag found")

    def validate_metadata_map_value(
        self, property_name: str, type_name: str, map_name: str, actual_value: str
    ) -> bool | str:
        """Validate whether the name found in an open metadata map property is valid.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        map_name: str
            The name of a map to validate.
        actual_value: str
             The actual value associated with the map to validate.

        Returns
        -------
        Bool - True, if validated.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_validate_metadata_map_value(
                property_name, type_name, map_name, actual_value
            )
        )
        return response

    async def _async_get_valid_metadata_value(
        self, property_name: str, type_name: str, preferred_value: str
    ) -> dict | str:
        """Retrieve details of a specific valid value for a property. Async version.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        preferred_value: str
            The preferred value of the property.

        Returns
        -------
        Dict if the value is found, otherwise an str indicating the value wasn't found.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        url = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/valid-metadata/get-value/"
            f"{property_name}?typeName={type_name}&preferredValue={preferred_value}"
        )

        response = await self._async_make_request("GET", url)
        return response.json().get("element", "No value found")

    def get_valid_metadata_value(
        self, property_name: str, type_name: str, preferred_value: str
    ) -> dict | str:
        """Retrieve details of a specific valid value for a property.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        preferred_value: str
            The preferred value of the property.

        Returns
        -------
        Dict if the value is found, otherwise an str indicating the value wasn't found.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_get_valid_metadata_value(
                property_name, type_name, preferred_value
            )
        )
        return response

    async def _async_get_valid_metadata_map_name(
        self, property_name: str, type_name: str, map_name: str
    ) -> dict | str:
        """Retrieve details of a specific valid name for a map property. Async version.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        map_name: str
            Map to return details of.

        Returns
        -------
        Dict if the value is found, otherwise an str indicating the value wasn't found.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        url = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/valid-metadata/get-map-name/"
            f"{property_name}?typeName={type_name}&mapName={map_name}"
        )

        response = await self._async_make_request("GET", url)
        return response.json().get("element", "No value found")

    def get_valid_metadata_map_name(
        self, property_name: str, type_name: str, map_name: str
    ) -> dict | str:
        """Retrieve details of a specific valid name for a map property.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        map_name: str
            Map to return details of.

        Returns
        -------
        Dict if the value is found, otherwise an str indicating the value wasn't found.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_get_valid_metadata_map_name(property_name, type_name, map_name)
        )
        return response

    async def _async_get_valid_metadata_map_value(
        self, property_name: str, type_name: str, preferred_value: str
    ) -> dict | str:
        """Retrieve details of a specific valid value for a map property. Async version.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        preferred_value: str
            Preferred value to return details of.

        Returns
        -------
        Dict if the value is found, otherwise an str indicating the value wasn't found.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        url = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/valid-metadata/get-map-value/"
            f"{property_name}?typeName={type_name}&preferredValue={preferred_value}"
        )

        response = await self._async_make_request("GET", url)
        return response.json().get("element", "No value found")

    def get_valid_metadata_map_value(
        self, property_name: str, type_name: str, preferred_value: str
    ) -> dict | str:
        """Retrieve details of a specific valid value for a map property.

        Parameters
        ----------
        property_name : str
            The name of the property to setup metadata map.
        type_name : str
            The type name of the property.
        preferred_value: str
            Preferred value to return details of.

        Returns
        -------
        Dict if the value is found, otherwise an str indicating the value wasn't found.

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_get_valid_metadata_map_value(
                property_name, type_name, preferred_value
            )
        )
        return response

    async def _async_get_valid_metadata_values(
        self,
        property_name: str,
        type_name: str = None,
        start_from: int = 0,
        page_size: int = None,
    ) -> list | str:
        """Retrieve list of values for the property. Async version.

        Parameters
        ----------
        property_name: str
            The property to query.
        type_name: str, opt
            The Open Metadata type to get the property values for. If not specified then all property values
            will be returned.
        start_from: int, opt
            Page to start from.
        page_size: int, opt
             Number of elements to return per page - if None, then default for class will be used.


        Returns
        -------
        List | str

        A list of collections linked off of the supplied element.

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        if page_size is None:
            page_size = self.page_size

        url = (
            f"{self.platform_url}/servers/{self.view_server}{self.valid_m_command_base}/get-valid-metadata-values/{property_name}"
            f"?typeName={type_name}&startFrom={start_from}&pageSize={page_size}"
        )

        resp = await self._async_make_request("GET", url)
        return resp.json().get("elementList", NO_ELEMENTS_FOUND)

    def get_valid_metadata_values(
        self,
        property_name: str,
        type_name: str = None,
        start_from: int = 0,
        page_size: int = None,
    ) -> list | str:
        """Retrieve list of values for the property.

        Parameters
        ----------
        property_name: str
            The property to query.
        type_name: str, opt
            The Open Metadata type to get the property values for. If not specified then all property values
            will be returned.
        start_from: int, opt
            Page to start from.
        page_size: int, opt
             Number of elements to return per page - if None, then default for class will be used.

        Returns
        -------
        List | str

        A list of collections linked off of the supplied element.

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """
        loop = asyncio.get_event_loop()
        resp = loop.run_until_complete(
            self._async_get_valid_metadata_values(
                property_name, type_name, start_from, page_size
            )
        )
        return resp

    async def _async_get_consistent_metadata_values(
        self,
        property_name: str,
        type_name: str,
        map_name: str,
        preferred_value: str,
        start_from: int = 0,
        page_size: int = None,
    ) -> list | str:
        """Retrieve all the consistent valid values for the requested property. Async version.

        Parameters
        ----------
        property_name : str
            The name of the property to retrieve the valid values for.
        type_name : str
            The open metadata type that the property is associated with.
        map_name : str
            A valid map name that associates a property with a value.
        preferred_value : str
            Preferred value to return details of.
        start_from: int, [default=0], optional
                    When multiple pages of results are available, the page number to start from.
        page_size: int, [default=None]
            The number of items to return in a single page. If not specified, the default will be taken from
            the class instance.

        Returns
        -------
        List | str

        A list of collections linked off of the supplied element.

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        if page_size is None:
            page_size = self.page_size

        url = (
            f"{self.platform_url}/servers/{self.view_server}{self.valid_m_command_base}/{property_name}/"
            f"consistent-metadata-values?typeName={type_name}&mapName={map_name}&preferredValue={preferred_value}"
            f"&startFrom={start_from}&pageSize={page_size}"
        )

        resp = await self._async_make_request("GET", url)
        return resp.json().get("elementList", NO_ELEMENTS_FOUND)

    def get_consistent_metadata_values(
        self,
        property_name: str,
        type_name: str,
        map_name: str,
        preferred_value: str,
        start_from: int = 0,
        page_size: int = None,
    ) -> list | str:
        """Retrieve all the consistent valid values for the requested property.

        Parameters
        ----------
        property_name : str
            The name of the property to retrieve the valid values for.
        type_name : str
            The open metadata type that the property is associated with.
        map_name : str
            A valid map name that associates a property with a value.
        preferred_value : str


        start_from: int, [default=0], optional
                    When multiple pages of results are available, the page number to start from.
        page_size: int, [default=None]
            The number of items to return in a single page. If not specified, the default will be taken from
            the class instance.
        Returns
        -------
        List | str

        A list of collections linked off of the supplied element.

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """
        loop = asyncio.get_event_loop()
        resp = loop.run_until_complete(
            self._async_get_consistent_metadata_values(
                property_name,
                type_name,
                map_name,
                preferred_value,
                start_from,
                page_size,
            )
        )
        return resp

    async def _async_set_consistent_metadata_values(
        self,
        property_name1: str,
        property_name2: str,
        type_name1: str,
        map_name1: str,
        preferred_value1: str,
        type_name2: str,
        map_name2: str,
        preferred_value2: str,
    ) -> None:
        """Set up consistent metadata values relationship between the two property values. Async version.

        Parameters
        ----------
        property_name1 : str
            The name of the first property.
        property_name2 : str
            The name of the second property.
        type_name1 : str
            The open metadata type that property1 is associated with.
        map_name1 : str
            First valid map name.
        preferred_value1 : str
            First preferred value.
        type_name2 : str
            The open metadata type that property2 is associated with.
        map_name2 : str
            Second valid map name.
        preferred_value2 : str
            Second preferred value.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        url = (
            f"{self.platform_url}/servers/{self.view_server}{self.valid_m_command_base}/{property_name1}/"
            f"consistent-metadata-values/{property_name2}?"
            f"typeName1={type_name1}&mapName1={map_name1}&preferredValue1={preferred_value1}&"
            f"typeName1={type_name2}&mapName2={map_name2}&preferredValue2={preferred_value2}"
        )

        await self._async_make_request("POST", url)
        return

    def set_consistent_metadata_values(
        self,
        property_name1: str,
        property_name2: str,
        type_name1: str,
        map_name1: str,
        preferred_value1: str,
        type_name2: str,
        map_name2: str,
        preferred_value2: str,
    ) -> None:
        """Set up consistent metadata values relationship between the two property values.

        Parameters
        ----------
        property_name1 : str
            The name of the first property.
        property_name2 : str
            The name of the second property.
        type_name1 : str
            The open metadata type that property1 is associated with.
        map_name1 : str
            First valid map name.
        preferred_value1 : str
            First preferred value.
        type_name2 : str
            The open metadata type that property2 is associated with.
        map_name2 : str
            Second valid map name.
        preferred_value2 : str
            Second preferred value.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_set_consistent_metadata_values(
                property_name1,
                property_name2,
                type_name1,
                map_name1,
                preferred_value1,
                type_name2,
                map_name2,
                preferred_value2,
            )
        )
        return

    #
    # Get all ...
    #
    async def _async_get_all_entity_types(self) -> list | str:
        """Returns the list of different types of metadata organized into two groups.  The first are the
            attribute type definitions (AttributeTypeDefs).  These provide types for attributes in full
            type definitions.  Full type definitions (TypeDefs) describe types for entities, relationships
            and classifications. Async version.

        Parameters
        ----------


        Returns
        -------
        List | str

            A list of all entity types.

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        url = f"{self.platform_url}/servers/{self.view_server}{self.valid_m_command_base}/open-metadata-types"

        resp = await self._async_make_request("GET", url)
        return resp.json()

    def get_all_entity_types(self) -> list | str:
        """Returns the list of different types of metadata organized into two groups.  The first are the
            attribute type definitions (AttributeTypeDefs).  These provide types for attributes in full
            type definitions.  Full type definitions (TypeDefs) describe types for entities, relationships
            and classifications. Async version.

        Parameters
        ----------


        Returns
        -------
        List | str

            A list of all entity types.

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """
        loop = asyncio.get_event_loop()
        resp = loop.run_until_complete(self._async_get_all_entity_types())
        return resp

    async def _async_get_all_entity_defs(self) -> list | str:
        """GReturns all the entity type definitions. Async version.

        Parameters
        ----------


        Returns
        -------
        List | str

            A list of all entity types.

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        url = f"{self.platform_url}/servers/{self.view_server}{self.valid_m_command_base}/open-metadata-types/entity-defs"

        resp = await self._async_make_request("GET", url)
        return resp.json().get("typeDefs", "No TypeDefs Found")

    def get_all_entity_defs(self) -> list | str:
        """Returns all the entity type definitions.

        Parameters
        ----------


        Returns
        -------
        List | str

            A list of all entity types.

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """
        loop = asyncio.get_event_loop()
        resp = loop.run_until_complete(self._async_get_all_entity_defs())
        return resp

    async def _async_get_all_relationship_defs(self) -> list | str:
        """Returns all the relationship type definitions. Async version.

        Parameters
        ----------


        Returns
        -------
        List | str

            A list of all entity types.

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        url = f"{self.platform_url}/servers/{self.view_server}{self.valid_m_command_base}/open-metadata-types/relationship-defs"

        resp = await self._async_make_request("GET", url)
        return resp.json().get("typeDefs", "No TypeDefs Found")

    def get_all_relationship_defs(self) -> list | str:
        """Returns all the relationship type definitions.

        Parameters
        ----------


        Returns
        -------
        List | str

            A list of all entity types.

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """
        loop = asyncio.get_event_loop()
        resp = loop.run_until_complete(self._async_get_all_relationship_defs())
        return resp

    async def _async_get_all_classification_defs(self) -> list | str:
        """Returns all the classification type definitions. Async version.

        Parameters
        ----------


        Returns
        -------
        List | str

            A list of all entity types.

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        url = f"{self.platform_url}/servers/{self.view_server}{self.valid_m_command_base}/open-metadata-types/classification-defs"

        resp = await self._async_make_request("GET", url)
        return resp.json().get("typeDefs", "No TypeDefs Found")

    def get_all_classification_defs(self) -> list | str:
        """Returns all the classification type definitions.

        Parameters
        ----------


        Returns
        -------
        List | str

            A list of all entity types.

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """
        loop = asyncio.get_event_loop()
        resp = loop.run_until_complete(self._async_get_all_classification_defs())
        return resp

    #
    # Get valid ...
    #

    async def _async_get_sub_types(self, type_name: str) -> list | str:
        """Returns all the TypeDefs for a specific subtype.  If a null result is returned it means the
            type has no subtypes. Async version.

        Parameters
        ----------
        type_name : str
            Type name to retrieve the sub-types for.

        Returns
        -------
        List | str

            A list of TypeDefs that can be attached to the specified type.

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        url = (
            f"{self.platform_url}/servers/{self.view_server}{self.valid_m_command_base}/open-metadata-types/sub-types/"
            f"{type_name}"
        )

        resp = await self._async_make_request("GET", url)
        return resp.json().get("typeDefs", "No TypeDefs Found")

    def get_sub_types(self, type_name: str) -> list | str:
        """Returns all the TypeDefs for a specific subtype.  If a null result is returned it means the
            type has no subtypes.

        Parameters
        ----------
        type_name : str
            Type name to retrieve the sub-types for.

        Returns
        -------
        List | str

            A list of TypeDefs that can be attached to the specified type.

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """
        loop = asyncio.get_event_loop()
        resp = loop.run_until_complete(
            self._async_get_valid_relationship_types(type_name)
        )
        return resp

    async def _async_get_valid_relationship_types(self, entity_type: str) -> list | str:
        """Returns all the TypeDefs for relationships that can be attached to the requested entity type.
            Async version.

        Parameters
        ----------
        entity_type : str
            The name of the entity type to retrieve the valid relationships for.


        Returns
        -------
        List | str

            A list of TypeDefs that can be attached to the specified entity type.

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        url = (
            f"{self.platform_url}/servers/{self.view_server}{self.valid_m_command_base}/open-metadata-types/{entity_type}/"
            f"attached-relationships"
        )

        resp = await self._async_make_request("GET", url)
        return resp.json().get("typeDefs", "No TypeDefs Found")

    def get_valid_relationship_types(self, entity_type: str) -> list | str:
        """Returns all the TypeDefs for relationships that can be attached to the requested entity type.
                    Async version.

            Parameters
            ----------
            entity_type : str
                 The name of the entity type to retrieve the valid relationships for.
             : str, optional
                The name of the server to  configure.
                If not provided, the server name associated with the instance is used.

            Returns
            -------
            List | str

                A list of TypeDefs that can be attached to the specified entity type.

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """
        loop = asyncio.get_event_loop()
        resp = loop.run_until_complete(
            self._async_get_valid_relationship_types(entity_type)
        )
        return resp

    async def _async_get_valid_classification_types(
        self, entity_type: str
    ) -> list | str:
        """Returns all the TypeDefs for classifications that can be attached to the requested entity type.
            Async version.

        Parameters
        ----------
        entity_type : str
            The name of the entity type to retrieve the classifications for.


        Returns
        -------
        List | str

            A list of classifications that can be attached to the specified entity type.

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        url = (
            f"{self.platform_url}/servers/{self.view_server}{self.valid_m_command_base}/open-metadata-types/{entity_type}/"
            f"attached-classifications"
        )

        resp = await self._async_make_request("GET", url)
        return resp.json().get("typeDefs", "No TypeDefs Found")

    def get_valid_classification_types(self, entity_type: str) -> list | str:
        """Returns all the TypeDefs for classifications that can be attached to the requested entity type.
                    Async version.

            Parameters
            ----------
            entity_type : str
                The name of the entity type to retrieve the classifications for.
             : str, optional
                The name of the server to  configure.
                If not provided, the server name associated with the instance is used.

            Returns
            -------
            List | str

                A list of classifications that can be attached to the specified entity type.

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """
        loop = asyncio.get_event_loop()
        resp = loop.run_until_complete(
            self._async_get_valid_classification_types(entity_type)
        )
        return resp

    async def _async_get_typedef_by_name(self, entity_type: str) -> dict | str:
        """Return the TypeDef identified by the unique name.
            Async version.

        Parameters
        ----------
        entity_type : str
            The name of the entity type to retrieve the typedef for.


        Returns
        -------
        dict | str

            The typedef associated with the type name

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """

        url = f"{self.platform_url}/servers/{self.view_server}{self.valid_m_command_base}/open-metadata-types/name/{entity_type}"

        resp = await self._async_make_request("GET", url)
        return resp.json().get("typeDef", "No TypeDefs Found")

    def get_typedef_by_name(self, entity_type: str) -> dict | str:
        """Return the TypeDef identified by the unique name.

        Parameters
        ----------
        entity_type : str
            The name of the entity type to retrieve the typedef for.


        Returns
        -------
        dict | str

            The typedef associated with the type name

        Raises
        ------

        PyegeriaInvalidParameterException
          If the client passes incorrect parameters on the request - such as bad URLs or invalid values
        PyegeriaAPIException
          Raised by the server when an issue arises in processing a valid request
        NotAuthorizedException
          The principle specified by the user_id does not have authorization for the requested action

        """
        loop = asyncio.get_event_loop()
        resp = loop.run_until_complete(self._async_get_typedef_by_name(entity_type))
        return resp


if __name__ == "__main__":
    print("Main-Valid Metadata Manager")



from datetime import datetime
import re
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from pyegeria.core.utils import (camel_to_title_case)
from markdown_it import MarkdownIt
from rich.console import Console
from loguru import logger
from pyegeria.core.config import settings

from pyegeria.view.mermaid_utilities import construct_mermaid_web
from pyegeria.view.base_report_formats import select_report_format, MD_SEPARATOR, get_report_spec_match
from pyegeria.models import to_camel_case

"""
Note on select_report_spec function:

This function and related data structures have been moved back to _output_formats.py.
Please import select_report_spec from pyegeria._output_formats instead of from this module.
"""

console = Console(width=settings.Environment.console_width)


def _extract_referenceable_properties(element: dict[str, Any]) -> dict[str, Any]:
    # Get general header attributes
    guid = element.get('elementHeader', {}).get("guid", None)
    if guid is None:
        return {}
    metadata_collection_id = element['elementHeader']['origin'].get("homeMetadataCollectionId", None)
    metadata_collection_name = element['elementHeader']['origin'].get("homeMetadataCollectionName", None)
    origin_category = element['elementHeader'].get("origin_category", None)
    created_by = element['elementHeader']["versions"].get("createdBy", None)
    create_time = element['elementHeader']["versions"].get("createTime", None)
    updated_by = element['elementHeader']["versions"].get("updatedBy", None)
    version = element['elementHeader']["versions"].get("version", None)
    type_name = element['elementHeader']["type"].get("typeName", None)
    classifications = element['elementHeader'].get("classifications", [])
    effective_from = element['elementHeader'].get("effectiveFrom", None)
    effective_to = element['elementHeader'].get("effectiveTo", None)

    # Get attributes from properties
    # properties = element['properties']
    # display_name = properties.get("name", "") or ""
    # if display_name == "":
    #     display_name = properties.get("displayName","")
    # description = properties.get("description", "") or ""
    # qualified_name = properties.get("qualifiedName", "") or ""
    # category = properties.get("category", "") or ""
    # version_identifier = properties.get("versionIdentifier", "") or ""
    # additional_properties = properties.get("additionalProperties", {}) or {}
    # extended_properties = properties.get("extendedProperties", {}) or {}
    #
    return {
        "GUID": guid,
        "metadata_collection_id": metadata_collection_id,
        "metadata_collection_name": metadata_collection_name,
        "origin_category": origin_category,
        "created_by": created_by,
        "create_time": create_time,
        "updated_by": updated_by,
        "version": version,
        "type_name": type_name,
        "classifications": classifications,

        # "display_name": display_name,
        # "description": description,
        # "qualified_name": qualified_name,
        # "category": category,
        # "version_identifier": version_identifier,
        # "additional_properties": additional_properties,
        # "extended_properties": extended_properties,
        "effective_from": effective_from,
        "effective_to": effective_to,
        }






def markdown_to_html(markdown_text: str) -> str:
    """
    Convert markdown text to HTML, with special handling for mermaid code blocks.

    Args:
        markdown_text: The markdown text to convert

    Returns:
        HTML string
    """
    # Initialize markdown-it
    md = MarkdownIt()

    # Find all mermaid code blocks
    mermaid_blocks = re.findall(r'```mermaid\n(.*?)\n```', markdown_text, re.DOTALL)

    # Replace each mermaid block with a placeholder
    placeholders = []
    for i, block in enumerate(mermaid_blocks):
        placeholder = f"MERMAID_PLACEHOLDER_{i}"
        markdown_text = markdown_text.replace(f"```mermaid\n{block}\n```", placeholder)
        placeholders.append((placeholder, block))

    # Convert markdown to HTML
    html_text = md.render(markdown_text)

    # Replace placeholders with rendered mermaid HTML
    for placeholder, mermaid_block in placeholders:
        mermaid_html = construct_mermaid_web(mermaid_block)
        html_text = html_text.replace(placeholder, mermaid_html)

    # Add basic HTML structure
    html_text = f"""
    <!DOCTYPE html>
    <html>
    <head>
        <meta charset="UTF-8">
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
        <title>Egeria Report</title>
        <style>
            body {{ font-family: Arial, sans-serif; line-height: 1.6; padding: 20px; }}
            h1 {{ color: #2c3e50; }}
            h2 {{ color: #3498db; }}
            pre {{ background-color: #f8f8f8; padding: 10px; border-radius: 5px; overflow-x: auto; }}
            table {{ border-collapse: collapse; width: 100%; margin-bottom: 20px; }}
            th, td {{ border: 1px solid #ddd; padding: 8px; text-align: left; }}
            th {{ background-color: #f2f2f2; }}
            tr:nth-child(even) {{ background-color: #f9f9f9; }}
        </style>
    </head>
    <body>
        {html_text}
    </body>
    </html>
    """

    return html_text

def make_preamble(obj_type: str, search_string: str, output_format: str = 'MD') -> Tuple[str, Optional[str]]:
    """
    Creates a preamble string and an elements action based on the given object type, search string,
    and output format.

    Args:
        obj_type: The type of object being updated or reported on (e.g., "Product", "Category").
        search_string: The search string used to filter objects. Defaults to "All Elements" if None.
        output_format: A format identifier determining the output structure.
            JSON - output standard json
            MD - output standard markdown with no preamble
            FORM - output markdown with a preamble for a form
            REPORT - output markdown with a preamble for a report

    Returns:
        tuple: A tuple containing:
            - A string representing the formatted update or report preamble.
            - A string or None indicating the action description for the elements,
              depending on the output format.
    """
    # search_string = search_string if search_string else "All Elements"
    elements_md = ""
    elements_action = "Update " + obj_type
    if output_format == "FORM":
        preamble = f"\n# Update {obj_type} Form - created at {datetime.now().strftime('%Y-%m-%d %H:%M')}\n"
        if search_string:
            preamble +=  f"\t {obj_type} found from the search string:  `{search_string}`\n\n"
        return preamble, elements_action
    elif output_format == "REPORT":
        elements_md += (f"# {obj_type} Report - created at {datetime.now().strftime('%Y-%m-%d %H:%M')}\n"
                       f"\t{obj_type}  found from the search string:  `{search_string}`\n\n")
        elements_action = None
        return elements_md, elements_action
    else:
        return "\n", elements_action

def make_md_attribute(attribute_name: str, attribute_value: str, output_type: str) -> Optional[str]:
    """
    Create a markdown attribute line for a given attribute name and value.

    Args:
        attribute_name: The name of the attribute
        attribute_value: The value of the attribute
        output_type: The output format (FORM, MD, REPORT)

    Returns:
        str: Formatted markdown for the attribute
    """
    output = ""
    if isinstance(attribute_value,str):
        attribute_value = attribute_value.strip() if attribute_value else ""
    elif isinstance(attribute_value,list) and len(attribute_value) > 0:
        attribute_value = ",\n".join(attribute_value)
    if attribute_name:
        if attribute_name.upper() == "GUID":
            attribute_title = attribute_name.upper()
        else:
            # attribute_title = attribute_name.title()
            attribute_title = camel_to_title_case(attribute_name)
    else:
        attribute_title = ""

    if output_type in ["FORM", "MD"]:
        if attribute_name.lower() in [ "mermaid", "solutionBlueprintMermaidGraph", "links", "implemented by", "sub_components"]:
            return '\n'

        output = f"## {attribute_title}\n{attribute_value}\n\n"
    elif output_type in ["REPORT", "MERMAID"]:
        if attribute_title in ['Mermaid Graph', 'Mermaid', 'Solution Blueprint Mermaid Graph']:
            output = f"## {attribute_title}\n\n```mermaid\n{attribute_value}\n```\n"
        elif attribute_value:
            output = f"## {attribute_title}\n{attribute_value}\n\n"
    return output

def format_for_markdown_table(text: str, guid: str = None) -> str:
    """
    Format text for markdown tables by replacing newlines with spaces and escaping pipe characters.
    No truncation is applied to allow full-length text display regardless of console width.

    Args:
        text (str): The text to format

    Returns:
        str: Formatted text safe for markdown tables
    """
    if not text:
        return ""
    # Replace newlines with spaces and escape pipe characters
    if isinstance(text, list):
        text = "\n".join(text)
    t = text.replace("\n", " ").replace("|", "\\|")
    if '::' in t and guid:
        t = f" [{t}](#{guid}) "
    return t


def populate_columns_from_properties(element: dict, columns_struct: dict) -> dict:
    """
    Populate a columns_struct with values from the element's properties.

    The element dict is expected to have a nested 'properties' dict whose keys are in camelCase.
    The columns_struct is expected to follow the format returned by select_report_spec, where
    columns are located at columns_struct['formats']['columns'] and each column is a dict containing
    at least a 'key' field expressed in snake_case. For each column whose snake_case key corresponds
    to a key in the element properties (after converting to camelCase), this function adds a 'value'
    entry to the column with the matching property's value.

    Args:
        element: The element containing a 'properties' dict with camelCase keys.
        columns_struct: The columns structure whose columns have snake_case 'key' fields.

    Returns:
        The updated columns_struct (the input structure is modified in place and also returned).
    """
    if not isinstance(columns_struct, dict):
        return columns_struct

    props = (element or {}).get('properties') or {}
    # If properties is not a dict, do nothing
    if not isinstance(props, dict):
        return columns_struct

    # Get the attributes list if present
    formats = columns_struct.get('formats') or {}
    columns = formats.get('attributes') if isinstance(formats, dict) else None
    if not isinstance(columns, list):
        return columns_struct

    for col in columns:
        try:
            key_snake = col.get('key') if isinstance(col, dict) else None
            if not key_snake:
                continue
            # Convert the snake_case key to camelCase to look up in properties
            key_camel = to_camel_case(key_snake)
            if key_camel in props:
                col['value'] = props.get(key_camel)
        except Exception as e:
            # Be resilient; log and continue
            logger.debug(f"populate_columns_from_properties: skipping column due to error: {e}")
            continue

    return columns_struct


def get_required_relationships(element: dict, columns_struct: dict) -> dict:
    """
    Populate relationship-derived column values in columns_struct based on top-level keys in the element.

    This function inspects the requested columns in columns_struct, converts each column key from
    snake_case to camelCase, and if a matching top-level key exists in the element, parses that value
    (typically lists of relationship beans) into a human-readable value (e.g., a comma-separated list
    of qualified names) and stores it under the column's 'value'. Columns not specified in the
    columns_struct are ignored. Existing non-empty 'value's are left as-is.

    Example: if a column with key 'member_of_collections' is present, this function will look for the
    top-level key 'memberOfCollections' in the element and derive a value if found.

    Args:
        element: The element dictionary containing top-level relationship lists (e.g., associatedGlossaries,
                 memberOfCollections, collectionMembers).
        columns_struct: The columns structure to augment with derived 'value's.

    Returns:
        The updated columns_struct (modified in place and returned).
    """
    if not isinstance(columns_struct, dict):
        return columns_struct

    formats = columns_struct.get('formats') or {}
    columns = formats.get('attributes') if isinstance(formats, dict) else None
    if not isinstance(columns, list):
        return columns_struct

    def _extract_name_from_item(item: Any) -> Optional[str]:
        """Best-effort extraction of a display/qualified name from a relationship item."""
        try:
            if isinstance(item, dict):
                # Common pattern: item['relatedElement']['properties']['qualifiedName']
                related = item.get('relatedElement') or item.get('related_element')
                if isinstance(related, dict):
                    props = related.get('properties') or {}
                    name = (
                        props.get('qualifiedName')
                        or props.get('displayName')
                        or props.get('name')
                    )
                    if name:
                        return name
                # Sometimes the properties may be at the top level of the item
                name = (
                    item.get('qualifiedName')
                    or item.get('displayName')
                    or item.get('name')
                )
                if name:
                    return name
            elif isinstance(item, str):
                return item
        except Exception as e:
            logger.debug(f"get_required_relationships: error extracting name from item: {e}")
        return None

    for col in columns:
        try:
            if not isinstance(col, dict):
                continue
            key_snake = col.get('key')
            if not key_snake:
                continue
            # If already has a non-empty value, don't overwrite
            if col.get('value') not in (None, ""):
                continue

            # Convert the snake_case key to camelCase to look up in top-level element
            key_camel = to_camel_case(key_snake)
            if key_camel not in element:
                continue

            top_val = element.get(key_camel)
            derived_value: str = ""
            if isinstance(top_val, list):
                names: List[str] = []
                for item in top_val:
                    nm = _extract_name_from_item(item)
                    if nm:
                        names.append(nm)
                derived_value = ", ".join(names)
            elif isinstance(top_val, dict):
                nm = _extract_name_from_item(top_val)
                derived_value = nm or ""
            else:
                # Primitive or unexpected type; coerce to string if not None
                derived_value = str(top_val) if top_val is not None else ""

            col['value'] = derived_value
        except Exception as e:
            logger.debug(f"get_required_relationships: skipping column due to error: {e}")
            continue

    return columns_struct


def generate_entity_md(elements: List[Dict], 
                      elements_action: str, 
                      output_format: str, 
                      entity_type: str,
                      extract_properties_func: Callable, 
                      get_additional_props_func: Optional[Callable] = None,
                       columns_struct: [dict] = None) -> str:
    """
    Generic method to generate markdown for entities.

    Args:
        elements (list): List of entity elements
        elements_action (str): Action description for elements
        output_format (str): Output format
        entity_type (str): Type of entity (Glossary, Term, Category, etc.)
        extract_properties_func: Function to extract properties from an element
        get_additional_props_func: Optional function to get additional properties
        columns (list): List of column name structures

    Returns:
        str: Markdown representation
    """
    heading = columns_struct.get("heading")
    if heading == "Default Base Attributes":
        elements_md = "## Reporting on Default Base Attributes - Perhaps couldn't find a valid combination of report_spec and output_format?\n\n"
    else:
        elements_md = ""
    base_columns = columns_struct['formats'].get('attributes') if columns_struct else None

    for element in elements:
        if element is None:
                continue
        guid = element.get('elementHeader', {}).get('guid')

        # Prefer new behavior: extractor returns an updated columns_struct with values
        returned_struct = None
        if columns_struct is not None:
            try:
                returned_struct = extract_properties_func(element, columns_struct)
            except TypeError:
                # Fallback for legacy extractors without columns_struct parameter
                returned_struct = None
        
        # Legacy fallback: get props dict if no columns_struct provided/returned
        props = {}
        if returned_struct is None:
            props = extract_properties_func(element) if callable(extract_properties_func) else {}

        # Get additional properties if function is provided
        additional_props = {}
        if get_additional_props_func:
            # Use guid if available, else try to get from props
            guid_for_fmt = guid or props.get('GUID')
            additional_props = get_additional_props_func(element, guid_for_fmt, output_format)

        # Determine display name
        display_name = None
        if returned_struct is not None:
            cols = returned_struct.get('formats', {}).get('attributes', [])
            # Find value from 'display_name' or 'title'
            for col in cols:
                if col.get('key') in ('display_name', 'title'):
                    display_name = col.get('value')
                    if display_name:
                        break
        else:
            display_name = props.get('display_name') or props.get('title')

        if display_name is None:
            display_name = "NO DISPLAY NAME"

        # Format header based on output format
        if output_format in ['FORM', 'MD']:
            elements_md += f"# {elements_action}\n\n"
            elements_md += f"## {entity_type} Name \n\n{display_name}\n\n"
        elif output_format == 'REPORT':
            elements_md += f'<a id="{(guid or props.get("GUID") or "No GUID" )}"></a>\n# {entity_type} Name: {display_name}\n\n'
        else:
            elements_md += f"## {entity_type} Name \n\n{display_name}\n\n"

        # Add attributes based on column spec if available, otherwise, add all (legacy)
        if returned_struct is not None:
            cols = returned_struct.get('formats', {}).get('attributes', [])
            for column in cols:
                name = column.get('name')
                key = column.get('key')
                value = column.get('value')
                if value in (None, "") and key in additional_props:
                    value = additional_props[key]
                if column.get('format'):
                    value = format_for_markdown_table(value, guid)
                elements_md += make_md_attribute(name, value, output_format)
            if wk := returned_struct.get("annotations", {}).get("wikilinks", None):
                elements_md += ", ".join(wk)
        elif base_columns:
            # If we have columns but extractor didn't return struct, use legacy props lookup
            for column in base_columns:
                key = column['key']
                name = column['name']
                value = ""
                if key in props:
                    value = props[key]
                elif key in additional_props:
                    value = additional_props[key]
                if column.get('format'):
                    value = format_for_markdown_table(value, guid or props.get('GUID'))
                elements_md += make_md_attribute(name, value, output_format)
            if wk := columns_struct.get("annotations", {}).get("wikilinks", None):
                elements_md += ", ".join(wk)
        else:
            # Legacy path without columns: dump all props
            for key, value in props.items():
                if output_format in ['FORM', 'MD', 'DICT'] and key == 'mermaid':
                    continue
                if key not in ['properties', 'display_name']:
                    if key == 'mermaid' and value == '':
                        continue
                    elements_md += make_md_attribute(key.replace('_', ' '), value, output_format)
            for key, value in additional_props.items():
                elements_md += make_md_attribute(key.replace('_', ' '), value, output_format)

        if element != elements[-1]:
            elements_md += MD_SEPARATOR

    return elements_md

def generate_entity_md_table(elements: List[Dict], 
                            search_string: str, 
                            entity_type: str, 
                            extract_properties_func: Callable,
                            columns_struct: dict,
                            get_additional_props_func: Optional[Callable] = None, 
                            output_format: str = 'LIST') -> str:
    """
    Generic method to generate a markdown table for entities.

    Args:
        elements (list): List of entity elements
        search_string (str): The search string used
        entity_type (str): Type of entity (Glossary, Term, Category, etc.)
        extract_properties_func: Function to extract properties from an element
        columns: List of column definitions, each containing 'name', 'key', and 'format' (optional)
        get_additional_props_func: Optional function to get additional properties
        output_format (str): Output format (FORM, REPORT, LIST, etc.)

    Returns:
        str: Markdown table
    """
    # Handle pluralization - if entity_type ends with 'y', use 'ies' instead of 's'
    target_type = columns_struct.get('target_type', entity_type)
    # if target_type.endswith('y'):
    #     target_type = target_type.replace('y', 'ies')
    # else:
    #     target_type = target_type.replace('s', 's')

    entity_type_plural = f"{target_type[:-1]}ies" if target_type.endswith('y') else f"{target_type}s"
    # entity_type_plural = target_type
    columns = columns_struct['formats'].get('attributes', [])
    heading = columns_struct.get("heading")
    if heading == "Default Base Attributes":
        elements_md = "## Reporting on Default Base Attributes - Perhaps couldn't find a valid combination of report_spec and output_format?\n\n"
    else:
        elements_md = ""

    if output_format == "LIST":
        elements_md = f"# {entity_type_plural} Table\n\n"
        elements_md += f"{entity_type_plural} found from the search string: `{search_string}`\n\n"

    # Add column headers
    header_row = "| "
    separator_row = "|"
    for column in columns:
        header_row += f"{column['name']} | "
        separator_row += "-------------|"

    elements_md += header_row + "\n"
    elements_md += separator_row + "\n"

    # Add rows
    for element in elements:
        guid = element.get('elementHeader', {}).get('guid', None)

        # Extractor returns columns_struct with values when possible
        try:
            returned_struct = extract_properties_func(element, columns_struct)
        except TypeError:
            returned_struct = None

        # For help mode, bypass extraction
        if output_format == "help":
            returned_struct = {"formats": {"attributes": columns}}

        # Additional props (if any)
        additional_props = {}
        if get_additional_props_func:
            additional_props = get_additional_props_func(element, guid, output_format)

        # Build row
        row = "| "
        if returned_struct is not None:
            for column in returned_struct.get('formats', {}).get('attributes', []):
                key = column.get('key')
                value = column.get('value')
                if (value in (None, "")) and key in additional_props:
                    value = additional_props[key]
                if column.get('format'):
                    value = format_for_markdown_table(value, guid)
                row += f"{value} | "
        else:
            # Legacy fallback: read from props dict
            props = extract_properties_func(element)
            for column in columns:
                key = column['key']
                value = ""
                if key in props:
                    value = props[key]
                elif key in additional_props:
                    value = additional_props[key]
                if column.get('format'):
                    value = format_for_markdown_table(value, guid or props.get('GUID'))
                row += f"{value} | "

        elements_md += row + "\n"
        # if wk := columns_struct.get("annotations",{}).get("wikilinks", None):
        #     elements_md += ", ".join(wk)
    return elements_md

def generate_entity_dict(elements: List[Dict], 
                        extract_properties_func: Callable, 
                        get_additional_props_func: Optional[Callable] = None,
                        include_keys: Optional[List[str]] = None, 
                        exclude_keys: Optional[List[str]] = None,
                        columns_struct: dict = None,
                        output_format: str = 'DICT') -> List[Dict]:
    """
    Generic method to generate a dictionary representation of entities.

    Args:
        elements (list): List of entity elements
        extract_properties_func: Function to extract properties from an element
        get_additional_props_func: Optional function to get additional properties
        include_keys: Optional list of keys to include in the result (if None, include all)
        exclude_keys: Optional list of keys to exclude from the result (if None, exclude none)
        columns_struct: Optional dict of columns to include (if None, include all)
        output_format (str): Output format (FORM, REPORT, DICT, etc.)

    Returns:
        list: List of entity dictionaries
    """
    result = []

    #####
    # Add attributes based on column spec if available, otherwise, add all
    for element in elements:
        if element is None:
            continue

        guid = element.get('elementHeader', {}).get('guid')

        returned_struct = None
        if columns_struct is not None:
            try:
                returned_struct = extract_properties_func(element, columns_struct)
            except TypeError as e:
                logger.info(f"Error - didn't find extractor?: {e}")
                returned_struct = None

        # Get additional properties if function is provided
        additional_props = {}
        if get_additional_props_func:
            additional_props = get_additional_props_func(element, guid, output_format)

        # Create entity dictionary
        entity_dict = {}

        columns = columns_struct['formats'].get('attributes', None) if columns_struct else None
        if returned_struct is not None:
            for column in returned_struct.get('formats', {}).get('attributes', []):
                key = column.get('key')
                name = column.get('name')
                value = column.get('value')
                if (value in (None, "")) and key in additional_props:
                    value = additional_props[key]
                if column.get('format'):
                    value = format_for_markdown_table(value, guid)
                # Use machine key for DICT output to avoid name collisions (e.g., multiple 'GUID' display names)
                dict_key = key or name
                # If the key already exists and differs only by case, preserve both by preferring exact key
                if dict_key in entity_dict and name and name != dict_key:
                    # suffix with _1 to avoid overwrite
                    suffix_idx = 1
                    tmp_key = f"{dict_key}_{suffix_idx}"
                    while tmp_key in entity_dict:
                        suffix_idx += 1
                        tmp_key = f"{dict_key}_{suffix_idx}"
                    dict_key = tmp_key
                entity_dict[dict_key] = value
        elif columns:
            for column in columns:
                key = column['key']
                name = column['name']
                value = ""
                props = extract_properties_func(element)
                if key in props:
                    value = props[key]
                elif key in additional_props:
                    value = additional_props[key]
                if  column.get('format', None):
                    value = format_for_markdown_table(value, guid or props.get('GUID'))
                dict_key = key or name
                if dict_key in entity_dict and name and name != dict_key:
                    suffix_idx = 1
                    tmp_key = f"{dict_key}_{suffix_idx}"
                    while tmp_key in entity_dict:
                        suffix_idx += 1
                        tmp_key = f"{dict_key}_{suffix_idx}"
                    dict_key = tmp_key
                entity_dict[dict_key] = value
        else:
            props = extract_properties_func(element, columns_struct)
            # Add properties based on include/exclude lists
            for key, value in props.items():
                if key not in ['properties', 'mermaid']:  # Skip the raw properties object
                    if (include_keys is None or key in include_keys) and (
                            exclude_keys is None or key not in exclude_keys):
                        entity_dict[key] = value

            # Add additional properties
            for key, value in additional_props.items():
                if (include_keys is None or key in include_keys) and (exclude_keys is None or key not in exclude_keys):
                    entity_dict[key] = value

        result.append(entity_dict)
    #####
    # for element in elements:
    #     if element is None:
    #         continue
    #     props = extract_properties_func(element)
    #
    #     # Get additional properties if function is provided
    #     additional_props = {}
    #     if get_additional_props_func:
    #         additional_props = get_additional_props_func(element,props['GUID'], output_format)
    #
    #     # Create entity dictionary
    #     entity_dict = {}
    #
    #     # Add properties based on include/exclude lists
    #     for key, value in props.items():
    #         if key not in [ 'properties', 'mermaid']:  # Skip the raw properties object
    #             if (include_keys is None or key in include_keys) and (
    #                     exclude_keys is None or key not in exclude_keys):
    #                 entity_dict[key] = value
    #
    #     # Add additional properties
    #     for key, value in additional_props.items():
    #         if (include_keys is None or key in include_keys) and (exclude_keys is None or key not in exclude_keys):
    #             entity_dict[key] = value
    #
    #     result.append(entity_dict)

    return result

def resolve_output_formats(entity_type: str,
                           output_format: str,
                           report_spec: Optional[Union[str, dict]] = None,
                           default_label: Optional[str] = None,
                           **kwargs) -> Optional[dict]:
    """
    Resolve a report format structure given an entity type, the desired output format
    (e.g., DICT, LIST, MD, REPORT, FORM), and either a label (str) or a dict of format sets.

    Backward compatibility:
    - Accepts legacy kwarg 'report_spec' and treats it as report_spec.

    Selection order:
    - If report_spec is a str: select by label.
    - If report_spec is a dict: use get_report_spec_match to pick a matching format.
    - Else: try selecting by entity_type or default_label.
    - Fallback: select "Default".
    """
    from pyegeria.view.base_report_formats import get_report_spec_match

    if report_spec is None and isinstance(kwargs, dict):
        if 'report_spec' in kwargs:
            report_spec = kwargs.get('report_spec')
        elif 'report_format' in kwargs:
            report_spec = kwargs.get('report_format')
        elif 'output_format_spec' in kwargs:
            report_spec = kwargs.get('output_format_spec')

    if isinstance(report_spec, str):
        return select_report_format(report_spec, output_format)
    if isinstance(report_spec, dict):
        return get_report_spec_match(report_spec, output_format)

    label = default_label or entity_type
    fmt = select_report_format(label, output_format)
    if fmt is None:
        fmt = select_report_format("Default", output_format)
    return fmt


def overlay_additional_values(columns_struct: dict, extra: Optional[dict]) -> dict:
    """
    Overlay extra values into columns_struct only where the column's value is empty or missing.
    Returns the modified columns_struct.
    """
    if not isinstance(columns_struct, dict) or not extra:
        return columns_struct
    columns = columns_struct.get('formats', {}).get('attributes')
    if not isinstance(columns, list):
        return columns_struct
    for col in columns:
        if not isinstance(col, dict):
            continue
        key = col.get('key')
        if not key:
            continue
        if col.get('value') in (None, "") and key in extra:
            col['value'] = extra[key]
    return columns_struct


def populate_common_columns(
    element: dict,
    columns_struct: dict,
    *,
    include_header: bool = True,
    include_relationships: bool = True,
    include_subject_area: bool = True,
    mermaid_source_key: str = 'mermaidGraph',
    mermaid_dest_key: str = 'mermaid'
) -> dict:
    """
    Populate the common columns in columns_struct based on a standard Egeria element shape.

    Steps:
    - Populate from element.properties (camelCase mapped from snake_case keys)
    - Optionally overlay header-derived values (GUID, type_name, times, etc.)
    - Optionally populate relationship-based columns via get_required_relationships
    - Optionally populate subject_area from element.elementHeader.subjectArea.classificationProperties.subjectAreaName
    - If a column with key == mermaid_dest_key is present, set it from mermaid_source_key
    - Do not overwrite non-empty values already set
    """
    # 1) Base properties
    col_data = populate_columns_from_properties(element, columns_struct)
    columns_list = col_data.get('formats', {}).get('attributes', [])

    # 2) Header overlay
    header_props = _extract_referenceable_properties(element) if include_header else {}
    guid = header_props.get('GUID') if include_header else None
    if include_header:
        for column in columns_list:
            if not isinstance(column, dict):
                continue
            key = column.get('key')
            if not key:
                continue
            if column.get('value') not in (None, ""):
                continue
            if key in header_props:
                column['value'] = header_props.get(key)
            elif isinstance(key, str) and key.lower() == 'guid':
                column['value'] = guid

    # 3) Relationships
    if include_relationships:
        col_data = get_required_relationships(element, col_data)

    # 4) Subject area
    if include_subject_area:
        try:
            subject_area = element.get('elementHeader', {}).get('subjectArea') or ""
            subj_val = ""
            if isinstance(subject_area, dict):
                subj_val = subject_area.get('classificationProperties', {}).get('subjectAreaName', '')
            for column in columns_list:
                if column.get('key') == 'subject_area' and column.get('value') in (None, ""):
                    column['value'] = subj_val
        except Exception as e:
            logger.debug(f"populate_common_columns: subject_area handling error: {e}")

    # 5) Mermaid
    try:
        mermaid_val = element.get(mermaid_source_key, '') or ''
        for column in columns_list:
            if column.get('key') == mermaid_dest_key and column.get('value') in (None, ""):
                column['value'] = mermaid_val
                break
    except Exception as e:
        logger.debug(f"populate_common_columns: mermaid handling error: {e}")

    return col_data


def extract_mermaid_only(elements: Union[Dict, List[Dict]]) -> Union[str, List[str]]:
    """
    Extract mermaid graph data from elements.

    Args:
        elements: Dictionary or list of dictionaries containing element data

    Returns:
        String or list of strings containing mermaid graph data
    """
    if isinstance(elements, dict):
        mer = elements.get('mermaidGraph', None)
        if mer:
            return f"\n```mermaid\n{mer}\n```"
        else:
            return "---"


    result = []
    for element in elements:
        mer = element.get('mermaidGraph', "---")
        mer_out = f"\n\n```mermaid\n{mer}\n```" if mer else "---"
        result.append(mer_out)
    return result

def extract_basic_dict(elements: Union[Dict, List[Dict]]) -> Union[Dict, List[Dict]]:
    """
    Extract basic dictionary data from elements.

    Args:
        elements: Dictionary or list of dictionaries containing element data

    Returns:
        Dictionary or list of dictionaries with extracted data
    """
    if isinstance(elements, dict):
        body = {'guid': elements['elementHeader']['guid']}
        for key in elements['properties']:
            body[key] = elements['properties'][key]

        # Add classifications if present
        classifications = elements['elementHeader'].get('classifications', [])
        if classifications:
            classification_names = "["
            for classification in classifications:
                if len(classification_names) > 1:
                    classification_names += ", "
                classification_names += f"{classification['classificationName']}"
            body['classification_names'] = classification_names + ']'

        return body

    result = []
    for element in elements:
        if element is None:
            continue
        body = {'guid': element['elementHeader']['guid']}
        for key in element['properties']:
            body[key] = element['properties'][key]

        # Add classifications if present
        classifications = element['elementHeader'].get('classifications', [])
        if classifications:
            classification_names = "["
            for classification in classifications:
                if len(classification_names) > 1:
                    classification_names += ", "
                classification_names += f"{classification['classificationName']}"
            body['classifications'] = classification_names + ']'

        result.append(body)
    return result

def _extract_default_properties(self, element: dict, columns_struct: dict) -> dict:
    props = element.get('properties', {}) or {}
    normalized = {
        'properties': props,
        'elementHeader': element.get('elementHeader', {}),
    }
    # Common population pipeline
    col_data = populate_common_columns(element, columns_struct)
    columns_list = col_data.get('formats', {}).get('attributes', [])

    return col_data


def _generate_default_output(self, elements: dict | list[dict], search_string: str,
                             element_type_name: str | None,
                             output_format: str = 'DICT',
                             report_format: dict | str | None = None,
                             **kwargs) -> str | list[dict]:
    entity_type = 'Referenceable'
    # Backward compatibility: accept legacy kwarg
    if report_format is None and isinstance(kwargs, dict) and 'report_spec' in kwargs:
        report_format = kwargs.get('report_spec')
    if report_format:
        if isinstance(report_format, str):
            output_formats = select_report_format(report_format, output_format)
        elif isinstance(report_format, dict):
            output_formats = get_report_spec_match(report_format, output_format)
        else:
            output_formats = None
    else:
        output_formats = select_report_format(entity_type, output_format)
    if output_formats is None:
        output_formats = select_report_format('Default', output_format)
    return generate_output(
        elements=elements,
        search_string=search_string,
        entity_type=entity_type,
        output_format=output_format,
        extract_properties_func=_extract_default_properties,
        get_additional_props_func=None,
        columns_struct=output_formats,
    )


def generate_output(elements: Union[Dict, List[Dict]],
               search_string: str,
               entity_type: str,
               output_format: str,
               extract_properties_func: Callable,
               get_additional_props_func: Optional[Callable] = None,
               columns_struct: dict = None) -> Union[str, list[dict]]:
    """
    Generate output in the specified format for the given elements.

    Args:
        elements: Dictionary or list of dictionaries containing element data
        search_string: The search string used to find the elements
        entity_type: The type of entity (e.g., "Glossary", "Term", "Category")
        output_format: The desired output format (MD, FORM, REPORT, LIST, DICT, MERMAID, HTML)
        extract_properties_func: Function to extract properties from an element
        get_additional_props_func: Optional function to get additional properties
        columns: Optional list of column definitions for table output

    Returns:
        Formatted output as string or list of dictionaries
    """
    columns = columns_struct['formats'].get('attributes',None) if columns_struct else None
    if not columns:
        columns_struct = select_report_format("Default",output_format)
        if columns_struct:
            columns = columns_struct.get('formats', {}).get('attributes', None)

    target_type = columns_struct.get('target_type', entity_type) if columns_struct else entity_type
    if target_type is None:
        target_type = entity_type

    # Ensure elements is a list
    if isinstance(elements, dict):
        elements = [elements]

    # Handle empty search string
    if search_string is None or search_string == '':
        search_string = "All"

    # Set the output format to DICT to return values to table display
    if output_format == "TABLE":
        output_format = "DICT"

    # Generate output based on format
    if output_format == 'MERMAID':
        return extract_mermaid_only(elements)

    elif output_format == 'HTML':
        # First generate the REPORT format output
        report_output = generate_output(
            elements=elements,
            search_string=search_string,
            entity_type=entity_type,
            output_format="REPORT",
            extract_properties_func=extract_properties_func,
            get_additional_props_func=get_additional_props_func,
            columns_struct=columns_struct
        )

        # Convert the markdown to HTML
        return markdown_to_html(report_output)

    elif output_format == 'DICT':
        return generate_entity_dict(
            elements=elements,
            extract_properties_func=extract_properties_func,
            get_additional_props_func=get_additional_props_func,
            exclude_keys=['properties'],
            columns_struct=columns_struct,
            output_format=output_format
        )

    elif output_format == 'LIST':
        if columns is None:
            raise ValueError("Columns must be provided for LIST output format")

        return generate_entity_md_table(
            elements=elements,
            search_string=search_string,
            entity_type=entity_type,
            extract_properties_func=extract_properties_func,
            columns_struct=columns_struct,
            get_additional_props_func=get_additional_props_func,
            output_format=output_format
        )

    else:  #  MD, FORM, REPORT
        elements_md, elements_action = make_preamble(
            obj_type=target_type,
            search_string=search_string,
            output_format=output_format
        )

        elements_md += generate_entity_md(
            elements=elements,
            elements_action=elements_action,
            output_format=output_format,
            entity_type=target_type,
            extract_properties_func=extract_properties_func,
            get_additional_props_func=get_additional_props_func,
            columns_struct = columns_struct
        )

        return elements_md

# Machine keys explanation
# -------------------------
# In pyegeria’s reporting system, each output column is defined with two identifiers:
# - name: the human-readable display label (e.g., "GUID", "Display Name") shown in tables and reports.
# - key: the stable, programmatic identifier (usually snake_case, e.g., guid, display_name) used internally.
#
# We refer to the key as the machine key. It is intended to be:
# - Unique within a given row/spec: so downstream code can unambiguously reference values.
# - Stable over time: it should not change with presentation tweaks or localization.
# - Program-friendly: lowercase with underscores, matching the data model or derivation (e.g., header vs properties).
#
# Why DICT output uses machine keys
# - Display names are for humans and may repeat across different sources of data in one row. For example,
#   Collections may include a GUID from the element header (key: guid, display name: "GUID") and a related member GUID
#   (key: GUID, display name: "GUID"). If the DICT used display names, one would overwrite the other.
# - By using the machine key (column['key']) as the dictionary key, DICT output avoids these collisions and remains
#   deterministic for programmatic consumers.
#
# Collision handling
# - If, within an unusual spec, two columns share the same key, DICT generation adds a numeric suffix (e.g., guid_1)
#   to preserve both values. This is a safeguard and such collisions should be avoided in specs.
#
# Guidance for spec authors
# - Choose clear, stable machine keys in snake_case that map to actual data fields or well-defined derivations.
# - Reserve display names for presentation; do not rely on them for programmatic access.
# - When you need two visually identical labels (e.g., both "GUID"), ensure their keys differ (e.g., guid vs GUID) to
#   reflect different sources/semantics.



#!/usr/bin/env python3
"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

Unit tests for the Utils helper functions using the Pytest framework.


A simple display for my profile
"""
import argparse
import os
import time

from rich import box
from rich.console import Console
from rich.table import Table

from pyegeria import (
    RegisteredInfo,
    print_basic_exception,
    PyegeriaException,
    settings, load_app_config, pretty_print_config,
    config_logging,
    save_mermaid_html,
)

app_config = settings.Environment
config_path = os.path.join(app_config.pyegeria_config_directory, app_config.pyegeria_config_file)

EGERIA_USER = os.environ.get("EGERIA_USER", "erinoverview")
EGERIA_USER_PASSWORD = os.environ.get("EGERIA_USER_PASSWORD", "secret")
EGERIA_MERMAID_FOLDER = os.path.join(app_config.pyegeria_root, app_config.egeria_mermaid_folder)
conf = load_app_config(config_path)
# print(f"Loading config from {config_path} and mermaid folder is {EGERIA_MERMAID_FOLDER}")
console = Console(width=app_config.console_width)
config_logging()


def display_asset_types(
    server: str,
    url: str,
    username: str,
    user_password: str,
    jupyter: bool = app_config.egeria_jupyter,
    width: int = app_config.console_width,
):
    r_client = RegisteredInfo(server, url, username)
    token = r_client.create_egeria_bearer_token(username, user_password)
    asset_types = r_client.list_asset_types()

    def generate_table() -> Table:
        """Make a new table."""
        table = Table(
            title=f"Asset Types for: {url} @ {time.asctime()}",
            style="bold bright_white on black",
            row_styles=["bold bright_white on black"],
            header_style="white on dark_blue",
            title_style="bold white on black",
            show_lines=True,
            box=box.ROUNDED,
            caption=f"Asset Types from Server '{server}' @ Platform - {url}",
            expand=True,
        )

        table.add_column("Name")
        table.add_column("Description")
        table.add_column("SuperType")
        table.add_column("Version")

        name = " "
        description = " "
        version = " "
        super_type = " "
        if len(asset_types) > 0:
            for a_type in asset_types:
                name = a_type.get("name", "none")
                description = a_type.get("description", "none")
                version = a_type.get("version", " ")
                super_type = a_type.get("superType", "none")
                table.add_row(name, description, super_type, str(version))
        return table

    try:
        console = Console(width=width, force_terminal=not jupyter)
        with console.pager(styles=True):
            console.print(generate_table())

    except (
        PyegeriaException
    ) as e:
        print_basic_exception(e)
        assert e.related_http_code != "200", "Invalid parameters"
    finally:
        r_client.close_session()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--server", help="Name of the server to display status for")
    parser.add_argument("--url", help="URL Platform to connect to")
    parser.add_argument("--userid", help="User Id")
    parser.add_argument("--password", help="User Password")

    args = parser.parse_args()

    server = args.server if args.server is not None else app_config.egeria_view_server
    url = args.url if args.url is not None else app_config.egeria_platform_url
    userid = args.userid if args.userid is not None else EGERIA_USER
    user_pass = args.password if args.password is not None else EGERIA_USER_PASSWORD
    # guid = args.guid if args.guid is not None else None
    guid = None

    display_asset_types(server, url, userid, user_pass)


if __name__ == "__main__":
    main()



"""
This is an ongoing experiment in parsing and playing with Freddie docs
"""
import os
import sys
import uuid
from typing import Type
from datetime import datetime

from loguru import logger
from pydantic import ValidationError
from rich.console import Console
from rich.table import Table
from rich.box import Box
from rich.markdown import Markdown

import asyncio
from md_processing import (extract_command, process_provenance_command, get_current_datetime_string, command_list)
from md_processing.md_processing_utils.common_md_proc_utils import set_parse_summary_mode, set_usage_level
from md_processing.md_processing_utils.common_md_utils import set_attribute_log_level, ALL_GOVERNANCE_DEFINITIONS
from md_processing.md_processing_utils.md_processing_constants import PROJECT_SUBTYPES, COLLECTION_SUBTYPES
from md_processing.v1_legacy.command_mapping import setup_dispatcher
from md_processing.v2 import (
    UniversalExtractor, v2Dispatcher, AsyncBaseCommandProcessor,
    TermProcessor, TermRelationshipProcessor,
    DataCollectionProcessor, DataStructureProcessor, DataFieldProcessor, DataClassProcessor,
    BlueprintProcessor, ComponentProcessor, SupplyChainProcessor, SolutionLinkProcessor,
    ProjectProcessor, ProjectLinkProcessor,
    CollectionManagerProcessor, CSVElementProcessor, CollectionLinkProcessor,
    GovernanceProcessor, GovernanceLinkProcessor, GovernanceContextProcessor,
    FeedbackProcessor, TagProcessor, ExternalReferenceProcessor, FeedbackLinkProcessor
)

from pyegeria import EgeriaTech, PyegeriaException, print_basic_exception, print_validation_error
from pyegeria.core.config import settings

# Configure logging - module level default
log_format = "{time} | {level} | {function} | {line} | {message} | {extra}"

# Only configure logging if we are running as main
if __name__ == "__main__":
    logger.remove()
    logger.add(sys.stderr, level="INFO", format=log_format, colorize=True)
    logger.add("debug_log.log", rotation="1 day", retention="1 week", compression="zip", level="INFO", format=log_format,
               colorize=True)

# Load configuration from config/config.json with environment variable overrides
app_config = settings.Environment

# Get configuration values with environment variable fallbacks
EGERIA_METADATA_STORE = app_config.egeria_metadata_store
EGERIA_KAFKA_ENDPOINT = app_config.egeria_kafka_endpoint
EGERIA_PLATFORM_URL = app_config.egeria_platform_url
EGERIA_VIEW_SERVER = app_config.egeria_view_server
EGERIA_VIEW_SERVER_URL = app_config.egeria_view_server_url
EGERIA_INTEGRATION_DAEMON = app_config.egeria_integration_daemon
EGERIA_INTEGRATION_DAEMON_URL = app_config.egeria_integration_daemon_url
EGERIA_WIDTH = int(app_config.console_width or 220)
console = Console(width=EGERIA_WIDTH)
EGERIA_JUPYTER = bool(app_config.egeria_jupyter)
EGERIA_GLOSSARY_PATH = app_config.egeria_glossary_path
EGERIA_ROOT_PATH = app_config.pyegeria_root
EGERIA_INBOX_PATH = app_config.dr_egeria_inbox
EGERIA_OUTBOX_PATH = app_config.dr_egeria_outbox

# User credentials and admin credentials are only from environment variables (not stored in config for security)
EGERIA_ADMIN_USER = os.environ.get("ADMIN_USER", "garygeeke")
EGERIA_ADMIN_PASSWORD = os.environ.get("ADMIN_PASSWORD", "secret")
EGERIA_USER = os.environ.get("EGERIA_USER", "erinoverview")
EGERIA_USER_PASSWORD = os.environ.get("EGERIA_USER_PASSWORD", "secret")

# Legacy environment variables (deprecated, kept for backward compatibility)
EGERIA_HOME_GLOSSARY_GUID = os.environ.get("EGERIA_HOME_GLOSSARY_GUID", None)

async def process_md_file_v2(input_file: str, output_folder: str, directive: str, client: EgeriaTech, 
                            parse_summary: str = "none", attribute_logs: str = "debug",
                            usage_level: str = None) -> None:
    """
    Async processing path for Dr.Egeria v2.
    """
    if usage_level:
        set_usage_level(usage_level)
    set_parse_summary_mode(parse_summary)
    set_attribute_log_level(attribute_logs)

    expanded_input = os.path.abspath(os.path.expanduser(input_file))
    if os.path.exists(expanded_input):
        full_file_path = expanded_input
    elif EGERIA_INBOX_PATH in input_file:
        full_file_path = os.path.abspath(os.path.expanduser(os.path.join(EGERIA_ROOT_PATH, input_file)))
    else:
        full_file_path = os.path.abspath(os.path.expanduser(os.path.join(EGERIA_ROOT_PATH, EGERIA_INBOX_PATH, input_file)))

    logger.info(f"v2: Processing Markdown File: {full_file_path}")
    console.print(f"[cyan]v2: Processing Markdown File: {full_file_path}[/cyan]")

    if directive == "process":
        console.print("\n[bold cyan]*** INFO: PROCESS MODE ***[/bold cyan]")
        console.print("[cyan]Dr. Egeria will EXECUTE these commands and make PERMANENT CHANGES to Egeria.[/cyan]\n")
    
    try:
        with open(full_file_path, 'r') as f:
            content = f.read()
    except FileNotFoundError:
        console.print(f"[red]Error: File not found at path: {full_file_path}[/red]")
        return

    # 1. Extract commands using UniversalExtractor
    extractor = UniversalExtractor(content)
    commands = extractor.extract_commands()
    
    if not commands:
        logger.warning(f"No valid Egeria Markdown commands found in {full_file_path}")
        console.print(f"[bold yellow]Warning:[/bold yellow] No valid Egeria commands could be parsed from [cyan]'{input_file}'[/cyan].")
        console.print("Ensure your markdown file uses correctly formatted headers (e.g. `## Create Glossary`) that match known commands.")
        return
        
    # 2. Setup v2 Dispatcher
    dispatcher = v2Dispatcher(client)
    from md_processing.md_processing_utils.md_processing_constants import get_command_spec, build_command_variants

    def register_processor(base_command: str, processor_cls: Type[AsyncBaseCommandProcessor]):
        """Register a processor for a base command and all its variants/alternates."""
        spec = get_command_spec(base_command)
        if spec:
            variants = build_command_variants(base_command, spec)
            for variant in variants:
                dispatcher.register(variant, processor_cls)
        else:
            dispatcher.register(base_command, processor_cls)

    # Glossary
    register_processor("Create Glossary Term", TermProcessor)
    register_processor("Link Term-Term Relationship", TermRelationshipProcessor)

    # Data Designer
    register_processor("Create Data Specification", DataCollectionProcessor)
    register_processor("Create Data Dictionary", DataCollectionProcessor)
    register_processor("Create Data Structure", DataStructureProcessor)
    register_processor("Create Data Field", DataFieldProcessor)
    register_processor("Create Data Class", DataClassProcessor)
    register_processor("Link Data Field", DataFieldProcessor) # Uses same processor as fields usually? Or generic link?

    # Solution Architect
    register_processor("Create Solution Blueprint", BlueprintProcessor)
    register_processor("Create Solution Component", ComponentProcessor)
    register_processor("Create Information Supply Chain", SupplyChainProcessor)
    register_processor("Link Solution Component Peers", SolutionLinkProcessor)
    register_processor("Link Information Supply Chain Peers", SolutionLinkProcessor)
    register_processor("Link Information Supply Chain Segment", SolutionLinkProcessor)

    # Project
    register_processor("Create Project", ProjectProcessor)
    register_processor("Link Project Dependency", ProjectLinkProcessor)
    register_processor("Link Project Hierarchy", ProjectLinkProcessor)

    # Alternate Project Commands (Campaign, Task, etc.)
    for proj_type in PROJECT_SUBTYPES:
        register_processor(f"Create {proj_type}", ProjectProcessor)
        register_processor(f"Update {proj_type}", ProjectProcessor)

    # Collection Manager
    for coll_type in COLLECTION_SUBTYPES:
        register_processor(f"Create {coll_type}", CollectionManagerProcessor)
        register_processor(f"Update {coll_type}", CollectionManagerProcessor)

    register_processor("Create CSV Element", CSVElementProcessor)
    register_processor("Add Member to Collection", CollectionLinkProcessor)
    register_processor("Link Agreement Item", CollectionLinkProcessor)
    register_processor("Link Product Dependency", CollectionLinkProcessor)
    register_processor("Link Product-Product", CollectionLinkProcessor)
    register_processor("Attach Collection to Resource", CollectionLinkProcessor)
    register_processor("Link Digital Subscriber", CollectionLinkProcessor)

    # Governance
    from md_processing.md_processing_utils.md_processing_constants import GOV_COM_LIST
    for gov_cmd in GOV_COM_LIST:
        register_processor(gov_cmd, GovernanceProcessor)
    
    # Fallback to catch any from the list that might be missing from GOV_COM_LIST or use a different verb
    for gov_type in ALL_GOVERNANCE_DEFINITIONS + ["Data Lens"]:
        register_processor(f"Create {gov_type}", GovernanceProcessor)
        register_processor(f"Update {gov_type}", GovernanceProcessor)
    
    # Generic registration for gov links
    register_processor("Link Governance Peer", GovernanceLinkProcessor)
    register_processor("Detach Governance Peer", GovernanceLinkProcessor)
    register_processor("Link Governance Supporting", GovernanceLinkProcessor)
    register_processor("Detach Governance Supporting", GovernanceLinkProcessor)
    register_processor("Link Governed By", GovernanceLinkProcessor)
    register_processor("Detach Governed By", GovernanceLinkProcessor)
    register_processor("View Governance Definition Context", GovernanceContextProcessor)

    # Feedback / Tags / External References
    register_processor("Add Comment", FeedbackProcessor)
    register_processor("Update Comment", FeedbackProcessor)
    register_processor("Create Journal Entry", FeedbackProcessor)
    register_processor("Create Informal Tag", TagProcessor)
    register_processor("Update Informal Tag", TagProcessor)
    register_processor("Add Informal Tag", FeedbackLinkProcessor)
    register_processor("Detach Informal Tag", FeedbackLinkProcessor)
    register_processor("Link Tag->Element", FeedbackLinkProcessor)
    register_processor("Link Tag", FeedbackLinkProcessor)
    register_processor("Detach Tag", FeedbackLinkProcessor)
    register_processor("Create External Reference", ExternalReferenceProcessor)
    register_processor("Update External Reference", ExternalReferenceProcessor)
    register_processor("Link External Reference", FeedbackLinkProcessor)
    register_processor("Detach External Reference", FeedbackLinkProcessor)
    register_processor("Create Related Media", ExternalReferenceProcessor)
    register_processor("Update Related Media", ExternalReferenceProcessor)
    register_processor("Link Related Media", FeedbackLinkProcessor)
    register_processor("Detach Related Media", FeedbackLinkProcessor)
    register_processor("Create Cited Document", ExternalReferenceProcessor)
    register_processor("Update Cited Document", ExternalReferenceProcessor)
    register_processor("Link Cited Document", FeedbackLinkProcessor)
    register_processor("Detach Cited Document", FeedbackLinkProcessor)

    context = {
        "directive": directive,
        "input_file": input_file,
        "request_id": str(uuid.uuid4())
    }
    
    # 3. Execution (Parallel)
    # We dispatch all commands in the document concurrently.
    # Note: Complex documents with inter-command dependencies may require a more 
    # sophisticated strategy in the future.
    results = await dispatcher.dispatch_batch(commands, context)
    
    # 4. Handle results and display summary
    final_output = []
    
    # 4a. Print analysis to console (for both validate and process)
    for res in results:
        if res.get("analysis"):
            console.print(f"\n[bold blue]=== Command Analysis: {res.get('verb')} {res.get('object_type')} ===[/bold blue]")
            console.print(Markdown(res["analysis"]))
            console.print("\n")
    
    # Custom box style for dotted line row separators
    DOTTED = Box(
        "┏━┳┓\n"
        "┃ ┃┃\n"
        "┡━╇┩\n"
        "│ ││\n"
        "├·┼┤\n"
        "├─┼┤\n"
        "│ ││\n"
        "└─┴┘\n"
    )

    summary_table = Table(title="Dr. Egeria v2 Processing Summary", show_header=True, header_style="bold magenta", 
                          show_lines=True, box=DOTTED)
    summary_table.add_column("Command", style="cyan", overflow="fold")
    summary_table.add_column("Status", style="bold", overflow="fold")
    summary_table.add_column("Found", style="green", overflow="fold")
    summary_table.add_column("Display Name", style="blue", overflow="fold")
    summary_table.add_column("Qualified Name", style="magenta", overflow="fold")
    summary_table.add_column("Message", min_width=36, overflow="fold")

    updated = False
    prov_found = False
    for res in results:
        # Handle provenance section if encountered
        if res.get("verb") == "Provenance":
            prov_found = True
            final_output.append(process_provenance_command(input_file, ""))
        else:
            final_output.append(res["output"])

        # Only add actual commands to the summary table
        if not res.get("is_command", True):
            continue

        # Skip Provenance from summary table as it's a meta-command
        if res.get("verb") == "Provenance":
            continue

        status_color = "green" if res["status"] == "success" else "red" if res["status"] == "failure" else "yellow"
        found_str = "Yes" if res.get("found") else "No"
        summary_table.add_row(
            f"{res['verb']} {res['object_type']}",
            f"[{status_color}]{res['status'].upper()}[/{status_color}]",
            found_str,
            res.get("display_name", ""),
            res.get("qualified_name", ""),
            res["message"]
        )
        if res["status"] == "success" and directive == "process":
            updated = True

    # Add Provenance if not already present
    if not prov_found:
        final_output.append(process_provenance_command(input_file, ""))

    console.print("\n")
    console.print(summary_table)
    console.print("\n")

    has_warnings = any(res.get("warnings") for res in results)

    # Display warnings for both validate and process
    if has_warnings:
        console.print("[bold yellow]Processing Warnings:[/bold yellow]\n")
        for res in results:
            warnings = res.get("warnings", [])
            if warnings:
                console.print(f"[yellow]-- {res['verb']} {res['object_type']} ({res.get('qualified_name', 'Unknown')}) --[/yellow]")
                for w in warnings:
                    console.print(f"   [yellow]* {w}[/yellow]")
        console.print("\n")

    if directive == "display":
        from md_processing.md_processing_utils.common_md_utils import render_markdown
        title = "Validation Diagnosis Summary" if directive == "validate" else "Display Summary"
        console.print(f"[bold cyan]{title}:[/bold cyan]\n")
        for res in results:
            if res.get("output"):
                render_markdown(res["output"])
        console.print("\n")

    if directive == "validate":
        from md_processing.md_processing_utils.common_md_utils import render_markdown
        title = "Validation Diagnosis Summary" if directive == "validate" else "Display Summary"
        console.print(f"[bold cyan]{title}:[/bold cyan]\n")
        for res in results:
            # Skip non-command blocks (intro text, etc.) in validate/display mode
            if not res.get("is_command", True):
                continue

            if res.get("output"):
                render_markdown(res["output"])
        console.print("\n")
    if directive == "process" and updated:
        # Re-assemble the file (using joined outputs)
        content_to_write = "\n".join(final_output)
        
        path, filename = os.path.split(input_file)
        new_filename = f"processed-{get_current_datetime_string()}-{filename}"
        
        if output_folder:
            new_file_path = os.path.abspath(os.path.expanduser(os.path.join(EGERIA_ROOT_PATH, EGERIA_OUTBOX_PATH, output_folder, new_filename)))
        else:
            new_file_path = os.path.abspath(os.path.expanduser(os.path.join(EGERIA_ROOT_PATH, EGERIA_OUTBOX_PATH, new_filename)))
            
        os.makedirs(os.path.dirname(new_file_path), exist_ok=True)

        with open(new_file_path, 'w') as f2:
            f2.write(content_to_write)
            
        console.print(f"==> [bold green]SUCCESS[/bold green]: Output written to [blue]{new_file_path}[/blue]")
    elif directive == "process":
        console.print("[yellow]No updates detected. New File not created.[/yellow]")

    console.print(f"\n[bold green]v2: Processing complete for '{input_file}'[/bold green]")
    logger.info("v2: Processing complete")


@logger.catch
def process_md_file(input_file: str, output_folder: str, directive: str, server: str, url: str, userid: str,
                          user_pass: str, parse_summary: str = "none",
                          attribute_logs: str = "debug", usage_level: str = None) -> None:
    """
    Process a markdown file by parsing and executing Dr. Egeria md_commands. Write output to a new file.
    """
    if usage_level:
        set_usage_level(usage_level)
    use_v2 = os.environ.get("DR_EGERIA_V2", "true").lower() in ("true", "1", "yes")

    if use_v2:
        client = EgeriaTech(server, url, user_id=userid)
        client.create_egeria_bearer_token(userid, user_pass)
        asyncio.run(process_md_file_v2(input_file, output_folder, directive, client, parse_summary, attribute_logs, usage_level))
        return

    set_parse_summary_mode(parse_summary)
    set_attribute_log_level(attribute_logs)

    cmd_list = command_list
    # console = Console(width=int(EGERIA_WIDTH))
    client = EgeriaTech(server, url, user_id=userid)
    token = client.create_egeria_bearer_token(userid, user_pass)
    
    # Initialize the dispatcher
    dispatcher = setup_dispatcher()

    updated = False
    full_file_path = os.path.abspath(os.path.expanduser(os.path.join(EGERIA_ROOT_PATH, EGERIA_INBOX_PATH, input_file)))
    logger.info("\n\n====================================================\n\n")
    logger.info(f"Processing Markdown File: {full_file_path}")
    console.print(f"[cyan]Processing Markdown File: {full_file_path}[/cyan]")

    if directive != "validate":
        console.print("\n[bold cyan]*** INFO: PROCESS MODE ***[/bold cyan]")
        console.print("[cyan]Dr. Egeria will EXECUTE these commands and make PERMANENT CHANGES to Egeria.[/cyan]\n")
    try:
        with open(full_file_path, 'r') as f:
            lines = f.readlines()
    except FileNotFoundError:
        console.print(f"[red]Error: File not found at path: {full_file_path}[/red]")
        return {}  # Return empty dict if file not found

    final_output = []
    prov_found = False
    prov_output = (f"\n# Provenance\n\n* Results from processing file {input_file} on "
                   f"{datetime.now().strftime('%Y-%m-%d %H:%M')}\n")
    h1_blocks = []
    current_block = ""
    in_h1_block = False

    # Helper function to process the current block
    def process_current_block(current_block):
        nonlocal updated, final_output, prov_found, prov_output, h1_blocks, in_h1_block

        if not current_block:
            return  # No block to process

        potential_command = extract_command(current_block)  # Extract object_action
        if potential_command in cmd_list:
            # Process the block based on the object_action
            if potential_command == "Provenance":
                result = process_provenance_command(input_file, current_block)
                prov_found = True
            else:
                # Use the dispatcher for all other commands
                result = dispatcher.dispatch(potential_command, client, current_block, directive)

            # Error handling logic
            if result is None and potential_command != "Provenance":
                handler = dispatcher.get_handler(potential_command)
                if not handler:
                     print(f"\n===> Unknown command: {potential_command}")
                
            if result:
                if directive == "process":
                    updated = True
                    final_output.append(result)
                elif directive == "validate":
                    pass 
            elif directive == "process":
                # Only warn if it looked like a valid command but failed (or was unknown)
                print(f"\n==>\tErrors found while processing command: \'{potential_command}\'\n"
                      f"\tPlease correct and try again. \n")
                final_output.append(current_block)
                final_output.append('\n____\n')
        else:
            # If there is no object_action, append the block as-is
            final_output.append(current_block)

    # Main parsing loop
    for line in lines:
        line = line.strip()  # Remove leading/trailing whitespace

        # Handle a new H1 block (starting with `# `)
        if line.startswith("# "):
            if in_h1_block:
                # Process the current block before starting a new one
                process_current_block(current_block)

            # Start a new H1 block
            current_block = line
            in_h1_block = True

        # Handle the end of a block (line starts with `---`)
        elif line.startswith("___") or line.startswith("---"):
            if in_h1_block:
                # Process the current block when it ends with `---`
                current_block += f"\n{line}"
                process_current_block(current_block)
                current_block = ""  # Clear the block
                in_h1_block = False

        # Add lines to the current H1 block
        elif in_h1_block:
            current_block += f"\n{line}"

        # Append non-H1 content directly to the output
        else:
            final_output.append(line)

    # Ensure the final H1 block is processed if the file doesn't end with `---`
    if in_h1_block:
        process_current_block(current_block)

    # Join the final output list into a single string
    final_output = "\n".join(final_output) if isinstance(final_output, list) else final_output

    try:
        if updated:
            path, filename = os.path.split(input_file)  # Get both parts
            new_filename = f"processed-{get_current_datetime_string()}-{filename}"  # Create the new filename
            
            if output_folder:
                new_file_path = os.path.abspath(os.path.expanduser(os.path.join(EGERIA_ROOT_PATH, EGERIA_OUTBOX_PATH, output_folder, new_filename)))
            else:
                new_file_path = os.path.abspath(os.path.expanduser(os.path.join(EGERIA_ROOT_PATH, EGERIA_OUTBOX_PATH, new_filename)))
            os.makedirs(os.path.dirname(new_file_path), exist_ok=True)

            with open(new_file_path, 'w') as f2:
                f2.write(final_output)
                if not prov_found:
                    f2.write(prov_output)
            console.print(f"\n==> Output written to [blue]{new_file_path}[/blue]")
        else:
            if directive != 'display':
                console.print("\n[yellow]No updates detected. New File not created.[/yellow]")
                # logger.error(f"===> Unknown Command?  <===") 

    except PyegeriaException as e:
        print_basic_exception(e)
    except ValidationError as e:
        print_validation_error(e)

if __name__ == "__main__":
    import argparse
    import asyncio
    parser = argparse.ArgumentParser(description="Process Dr. Egeria Markdown files using v2.")
    parser.add_argument("--input-file", required=True, help="Input markdown file name (in Inbox)")
    parser.add_argument("--output-folder", default="", help="Optional output subfolder (in Outbox)")
    parser.add_argument("--directive", default="process", choices=["display", "validate", "process"], help="Action to perform")
    
    args = parser.parse_args()
    
    # Run the async v2 processor (mocking client for simple CLI run)
    client = EgeriaTech(settings.Environment.egeria_view_server, settings.Environment.egeria_view_server_url, user_id=os.environ.get("EGERIA_USER", "erinoverview"))
    asyncio.run(
        process_md_file_v2(
            args.input_file, 
            args.output_folder,
            args.directive,
            client
        )
    )
